# adt-cloud-worker
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""ADT Cloud worker functional code."""


from __future__ import print_function

import argparse
import configparser
import kombu
from kombu.log import get_logger
from kombu.mixins import ConsumerMixin
from kombu.utils.debug import setup_logging
import os
import socket
import subprocess
import tarfile
import tempfile

from adt_cloud_worker._cloud_support import (
    get_default_testbed_flavor,
    set_config_dict,
    upload_tarball_to_swift_container,
)
ROUTING_KEYS = (
    'amd64.nova',
    'i386.nova'
)
API_VERSION = "v1"

logger = get_logger(__name__)


class AdtNovaWorker(ConsumerMixin):

    def __init__(self, connection, config):
        self.connection = connection
        self.config = config
        # Set the nova configuration env vars:
        _set_nova_environment_variables(self.config.items('nova'))
        # Worker is named after the machine hostname created by juju:
        # juju-<env_name>-machine-<machine_number>
        # By getting rid of the '-machine' term, the worker name is
        # identical to its corresponding security-group.
        self.worker_name = socket.gethostname().replace('-machine', '')

    def get_consumers(self, Consumer, channel):
        """Return consumers instances for all configured queues."""
        exchange = kombu.Exchange("adt.exchange", type="topic")
        queues = [
            kombu.Queue(
                'adt.requests.{}.{}'.format(
                    key, API_VERSION), exchange, routing_key=key)
            for key in ROUTING_KEYS
        ]
        return [Consumer(queues=queues, callbacks=[self.process])]

    def process(self, body, message):
        """Process incomming test request.

        Run requested tests and posts results to the 'adt_results' queue
        for later checking.
        """
        logger.info('Got: {}'.format(body))
        body['worker'] = self.worker_name

        # Run requested tests safely.
        tarball_path = None
        try:
            _validate_request_contents(body)
            # Create and temporary directory for storing adt results.
            dir_prefix = 'adt-{}-{}'.format(
                self.worker_name, body['request_id'])
            with tempfile.TemporaryDirectory(prefix=dir_prefix) as result_dir:
                # Run `adt-run`.
                adt_kwargs = body.copy()
                adt_kwargs['result_dir'] = result_dir
                adt_kwargs['nova_extra_args'] = self.config.get(
                    'nova', 'extra_args').split()
                adt_kwargs['nova_flavor'] = get_default_testbed_flavor()
                arguments = _make_adt_argument_list(adt_kwargs)

                body['exit_code'] = run_adt(arguments)
                tarball_path = _create_tarball_from_directory(result_dir)
                # Upload the results tarball to swift with a temporary name
                # 'adt-<req_id>/results.tgz.tmp', result checker will rename
                # if after verification.
                container_name = 'adt-{}'.format(body['request_id'])
                upload_tarball_to_swift_container(tarball_path, container_name)
        except Exception as e:
            # Unexpected failures ...
            logger.error(e, exc_info=True)
            body['exit_code'] = '100'
        finally:
            # The post results
            # TODO: send error logging to result-checker in the message
            queue = self.connection.SimpleQueue('adt.results')
            queue.put(body)
            queue.close()
            if tarball_path and os.path.exists(tarball_path):
                os.remove(tarball_path)
            message.ack()
            logger.info('Done!')


def _validate_request_contents(request):
    """Validates that 'request' contains everything we need to process it.

    Returns None on success, raises KeyError on failure.

    """
    required_keys = {
        'package_name',
        'nova_image',
        'request_id',
    }
    missing = required_keys.difference(set(request.keys()))
    if missing:
        raise KeyError(
            "Request is missing the following required keys: %s" % ', '.join(
                sorted(list(missing)))
            )


def run_adt(arguments):
    """Run adt-run according to the given request parameters.

    Always resets the environment with configured nova variables.

    Returns the exit code from adt-run.
    """
    # TODO: We probably want something more clever here:
    try:
        subprocess.check_call(['adt-run'] + arguments)
    except subprocess.CalledProcessError as e:
        # log?
        # TODO: filter log content to avoid leaking cloud credentials.
        return e.returncode
    return 0


def _make_adt_argument_list(request_configuration):
    """Generate the argument list to adt-run, given the
    requestion_configuration dictionary.

    :returns: A tuple of arguments.
    """
    adt_run_args = [
        '-d',
        '--apt-source', request_configuration['package_name'],
        '--output-dir', request_configuration['result_dir'],
    ]
    if 'apt_pocket' in request_configuration:
        adt_run_args += [
            '--apt-pocket', request_configuration['apt_pocket'],
            '--apt-upgrade',
        ]
    adt_ssh_nova_args = [
        '--',
        '-d',
        '--flavor', request_configuration['nova_flavor'],
        '--image', request_configuration['nova_image'],
    ]
    if 'nova_extra_args' in request_configuration:
        adt_ssh_nova_args += request_configuration['nova_extra_args']

    return (
        adt_run_args +
        ['---', 'ssh', '-s', 'uci-nova', '-d'] +
        adt_ssh_nova_args
    )


def _set_nova_environment_variables(nova_config):
    """Set environment variables for nova configuration.

    'nova_config' will be read, and any keys that start with 'os_' will be
    set as environment variables.

    'nova_config' must be an iterable of (key, value) pairs.

    """
    # Setup nova environment variables based on the configuration file.
    for k, v in nova_config:
        if not k.startswith('os_'):
            continue
        os.environ[k.upper()] = str(v)


def _create_tarball_from_directory(directory_to_tar):
    """Create a gzip'd tarball of the given directory.

    :returns: the path to the newly created tarball. The caller is responsible
    for deleting the tarball.

    """
    # Compress the results directory in a temporary file.
    _unused, tarball_path = tempfile.mkstemp()
    dirname_for_tarball = os.path.basename(directory_to_tar)
    with tarfile.open(tarball_path, "w:gz") as tar:
        for entry in os.listdir(directory_to_tar):
            tar.add(
                os.path.join(directory_to_tar, entry),
                arcname=os.path.join(dirname_for_tarball, entry),
            )
    return tarball_path


def main():
    setup_logging(loglevel='DEBUG', loggers=[''])

    parser = argparse.ArgumentParser(
        description='ADT cloud worker ...')
    parser.add_argument('-c', '--conf', default='adt-service.conf',
                        help='Configuration file path')
    args = parser.parse_args()

    # Load configuration options.
    config = configparser.ConfigParser()
    config.read(args.conf)

    set_config_dict(config)

    amqp_uris = config.get('amqp', 'uris').split()

    with kombu.Connection(amqp_uris) as conn:
        try:
            worker = AdtNovaWorker(conn, config)
            worker.run()
        except KeyboardInterrupt:
            print('Bye!')


if __name__ == '__main__':
    main()
